@@ -49,15 +49,26 @@ var Ci = Components.interfaces;
 var Cc = Components.classes;
 var Cr = Components.results;
 
+const EXCLUDE_FROM_BACKUP_ANNO = "places/excludeFromBackup";
 const POST_DATA_ANNO = "bookmarkProperties/POSTData";
 const READ_ONLY_ANNO = "placesInternal/READ_ONLY";
 const LMANNO_FEEDURI = "livemark/feedURI";
 const LMANNO_SITEURI = "livemark/siteURI";
+const LMANNO_EXPIRATION = "livemark/expiration";
+const LMANNO_LOADFAILED = "livemark/loadfailed";
+const LMANNO_LOADING = "livemark/loading";
 
-//@line 62 "/mnt/Mozilla/build/toolkit/components/places/src/utils.js"
+// The RESTORE_*_NSIOBSERVER_TOPIC constants should match the #defines of the
+// same names in browser/components/places/src/nsPlacesImportExportService.cpp
+const RESTORE_BEGIN_NSIOBSERVER_TOPIC = "bookmarks-restore-begin";
+const RESTORE_SUCCESS_NSIOBSERVER_TOPIC = "bookmarks-restore-success";
+const RESTORE_FAILED_NSIOBSERVER_TOPIC = "bookmarks-restore-failed";
+const RESTORE_NSIOBSERVER_DATA = "json";
+
+//@line 73 "/Srcs/Xulrunner/build/toolkit/components/places/src/utils.js"
 // On other platforms, the transferable system converts "\r\n" to "\n".
 const NEWLINE = "\r\n";
-//@line 65 "/mnt/Mozilla/build/toolkit/components/places/src/utils.js"
+//@line 76 "/Srcs/Xulrunner/build/toolkit/components/places/src/utils.js"
 
 function QI_node(aNode, aIID) {
   var result = null;
@@ -213,10 +224,9 @@ var PlacesUtils = {
    * @returns true if the node is a visit item, false otherwise
    */
   nodeIsVisit: function PU_nodeIsVisit(aNode) {
-    const NHRN = Ci.nsINavHistoryResultNode;
     var type = aNode.type;
-    return type == NHRN.RESULT_TYPE_VISIT ||
-           type == NHRN.RESULT_TYPE_FULL_VISIT;
+    return type == Ci.nsINavHistoryResultNode.RESULT_TYPE_VISIT ||
+           type == Ci.nsINavHistoryResultNode.RESULT_TYPE_FULL_VISIT;
   },
 
   /**
@@ -250,8 +260,8 @@ var PlacesUtils = {
    * @returns true if the node is readonly, false otherwise
    */
   nodeIsReadOnly: function PU_nodeIsReadOnly(aNode) {
-    if (this.nodeIsFolder(aNode))
-      return this.bookmarks.getFolderReadonly(asQuery(aNode).folderItemId);
+    if (this.nodeIsFolder(aNode) || this.nodeIsDynamicContainer(aNode))
+      return this.bookmarks.getFolderReadonly(this.getConcreteItemId(aNode));
     if (this.nodeIsQuery(aNode) &&
         asQuery(aNode).queryOptions.resultType !=
           Ci.nsINavHistoryQueryOptions.RESULTS_AS_TAG_CONTENTS)
@@ -340,7 +350,7 @@ var PlacesUtils = {
    * @returns true if the node is a dynamic container item, false otherwise
    */
   nodeIsDynamicContainer: function PU_nodeIsDynamicContainer(aNode) {
-    if (aNode.type == NHRN.RESULT_TYPE_DYNAMIC_CONTAINER)
+    if (aNode.type == Ci.nsINavHistoryResultNode.RESULT_TYPE_DYNAMIC_CONTAINER)
       return true;
     return false;
   },
@@ -602,6 +612,10 @@ var PlacesUtils = {
         var parts = blob.split("\n");
         for (var i = 0; i < parts.length; i++) {
           var uriString = parts[i];
+          // text/uri-list is converted to TYPE_UNICODE but it could contain
+          // comments line prepended by #, we should skip them
+          if (uriString.substr(0, 1) == '\x23')
+            continue;
           // note: this._uri() will throw if uriString is not a valid URI
           if (uriString != "" && this._uri(uriString))
             nodes.push({ uri: uriString,
@@ -718,10 +732,15 @@ var PlacesUtils = {
    *        Array of objects, each containing the following properties:
    *        name, flags, expires, type, mimeType (only used for binary
    *        annotations) value.
+   *        If the value for an annotation is not set it will be removed.
    */
   setAnnotationsForURI: function PU_setAnnotationsForURI(aURI, aAnnos) {
     var annosvc = this.annotations;
     aAnnos.forEach(function(anno) {
+      if (!anno.value) {
+        annosvc.removePageAnnotation(aURI, anno.name);
+        return;
+      }
       var flags = ("flags" in anno) ? anno.flags : 0;
       var expires = ("expires" in anno) ?
         anno.expires : Ci.nsIAnnotationService.EXPIRE_NEVER;
@@ -743,10 +762,16 @@ var PlacesUtils = {
    *        Array of objects, each containing the following properties:
    *        name, flags, expires, type, mimeType (only used for binary
    *        annotations) value.
+   *        If the value for an annotation is not set it will be removed.
    */
   setAnnotationsForItem: function PU_setAnnotationsForItem(aItemId, aAnnos) {
     var annosvc = this.annotations;
+
     aAnnos.forEach(function(anno) {
+      if (!anno.value) {
+        annosvc.removeItemAnnotation(aItemId, anno.name);
+        return;
+      }
       var flags = ("flags" in anno) ? anno.flags : 0;
       var expires = ("expires" in anno) ?
         anno.expires : Ci.nsIAnnotationService.EXPIRE_NEVER;
@@ -762,18 +787,6 @@ var PlacesUtils = {
     });
   },
 
-  /**
-   * Helper for getting a serialized Places query for a particular folder.
-   * @param aFolderId The folder id to get a query for.
-   * @return string serialized place URI
-   */
-  getQueryStringForFolder: function PU_getQueryStringForFolder(aFolderId) {
-    var options = this.history.getNewQueryOptions();
-    var query = this.history.getNewQuery();
-    query.setFolders([aFolderId], 1);
-    return this.history.queriesToQueryString([query], 1, options);
-  },
-
   // identifier getters for special folders
   get placesRootId() {
     delete this.placesRootId;
@@ -800,6 +813,21 @@ var PlacesUtils = {
     return this.unfiledBookmarksFolderId = this.bookmarks.unfiledBookmarksFolder;
   },
 
+  /**
+   * Checks if aItemId is a root.
+   *
+   *   @param aItemId
+   *          item id to look for.
+   *   @returns true if aItemId is a root, false otherwise.
+   */
+  isRootItem: function PU_isRootItem(aItemId) {
+    return aItemId == PlacesUtils.bookmarksMenuFolderId ||
+           aItemId == PlacesUtils.toolbarFolderId ||
+           aItemId == PlacesUtils.unfiledBookmarksFolderId ||
+           aItemId == PlacesUtils.tagsFolderId ||
+           aItemId == PlacesUtils.placesRootId;
+  },
+
   /**
    * Set the POST data associated with a bookmark, if any.
    * Used by POST keywords.
@@ -909,84 +937,113 @@ var PlacesUtils = {
     return -1;
   },
 
-  // Returns true if a container has uris in its first level
-  // Has better performances than checking getURLsForContainerNode(node).length
+  /**
+   * Returns a nsNavHistoryContainerResultNode with forced excludeItems and
+   * expandQueries.
+   * @param   aNode
+   *          The node to convert
+   * @param   [optional] excludeItems
+   *          True to hide all items (individual bookmarks). This is used on
+   *          the left places pane so you just get a folder hierarchy.
+   * @param   [optional] expandQueries
+   *          True to make query items expand as new containers. For managing,
+   *          you want this to be false, for menus and such, you want this to
+   *          be true.
+   * @returns A nsINavHistoryContainerResultNode containing the unfiltered
+   *          contents of the container.
+   * @note    The returned container node could be open or closed, we don't
+   *          guarantee its status.
+   */
+  getContainerNodeWithOptions:
+  function PU_getContainerNodeWithOptions(aNode, aExcludeItems, aExpandQueries) {
+    if (!this.nodeIsContainer(aNode))
+      throw Cr.NS_ERROR_INVALID_ARG;
+
+    // excludeItems is inherited by child containers in an excludeItems view.
+    var excludeItems = asQuery(aNode).queryOptions.excludeItems ||
+                       asQuery(aNode.parentResult.root).queryOptions.excludeItems;
+    // expandQueries is inherited by child containers in an expandQueries view.
+    var expandQueries = asQuery(aNode).queryOptions.expandQueries &&
+                        asQuery(aNode.parentResult.root).queryOptions.expandQueries;
+
+    // If our options are exactly what we expect, directly return the node.
+    if (excludeItems == aExcludeItems && expandQueries == aExpandQueries)
+      return aNode;
+
+    // Otherwise, get contents manually.
+    var queries = {}, options = {};
+    this.history.queryStringToQueries(aNode.uri, queries, {}, options);
+    options.value.excludeItems = aExcludeItems;
+    options.value.expandQueries = aExpandQueries;
+    return this.history.executeQueries(queries.value,
+                                       queries.value.length,
+                                       options.value).root;
+  },
+
+  /**
+   * Returns true if a container has uri nodes in its first level.
+   * Has better performance than (getURLsForContainerNode(node).length > 0).
+   * @param aNode
+   *        The container node to search through.
+   * @returns true if the node contains uri nodes, false otherwise.
+   */
   hasChildURIs: function PU_hasChildURIs(aNode) {
     if (!this.nodeIsContainer(aNode))
       return false;
 
-    // in the Library left pane we use excludeItems
-    if (this.nodeIsFolder(aNode) && asQuery(aNode).queryOptions.excludeItems) {
-      var itemId = PlacesUtils.getConcreteItemId(aNode);
-      var contents = this.getFolderContents(itemId, false, false).root;
-      for (var i = 0; i < contents.childCount; ++i) {
-        var child = contents.getChild(i);
-        if (this.nodeIsURI(child))
-          return true;
-      }
-      return false;
+    var root = this.getContainerNodeWithOptions(aNode, false, true);
+    var oldViewer = root.parentResult.viewer;
+    var wasOpen = root.containerOpen;
+    if (!wasOpen) {
+      root.parentResult.viewer = null;
+      root.containerOpen = true;
     }
 
-    var wasOpen = aNode.containerOpen;
-    if (!wasOpen)
-      aNode.containerOpen = true;
     var found = false;
-    for (var i = 0; i < aNode.childCount && !found; i++) {
-      var child = aNode.getChild(i);
+    for (var i = 0; i < root.childCount && !found; i++) {
+      var child = root.getChild(i);
       if (this.nodeIsURI(child))
         found = true;
     }
-    if (!wasOpen)
-      aNode.containerOpen = false;
+
+    if (!wasOpen) {
+      root.containerOpen = false;
+      root.parentResult.viewer = oldViewer;
+    }
     return found;
   },
 
+  /**
+   * Returns an array containing all the uris in the first level of the
+   * passed in container.
+   * If you only need to know if the node contains uris, use hasChildURIs.
+   * @param aNode
+   *        The container node to search through
+   * @returns array of uris in the first level of the container.
+   */
   getURLsForContainerNode: function PU_getURLsForContainerNode(aNode) {
-    let urls = [];
-    if (this.nodeIsFolder(aNode) && asQuery(aNode).queryOptions.excludeItems) {
-      // grab manually
-      var itemId = this.getConcreteItemId(aNode);
-      let contents = this.getFolderContents(itemId, false, false).root;
-      for (let i = 0; i < contents.childCount; ++i) {
-        let child = contents.getChild(i);
-        if (this.nodeIsURI(child))
-          urls.push({uri: child.uri, isBookmark: this.nodeIsBookmark(child)});
-      }
-    }
-    else {
-      let result, oldViewer, wasOpen;
-      try {
-        let wasOpen = aNode.containerOpen;
-        result = aNode.parentResult;
-        oldViewer = result.viewer;
-        if (!wasOpen) {
-          result.viewer = null;
-          aNode.containerOpen = true;
-        }
-        for (let i = 0; i < aNode.childCount; ++i) {
-          // Include visible url nodes only
-          let child = aNode.getChild(i);
-          if (this.nodeIsURI(child)) {
-            // If the node contents is visible, add the uri only if its node is
-            // visible. Otherwise follow viewer's collapseDuplicates property,
-            // default to true
-            if ((wasOpen && oldViewer && child.viewIndex != -1) ||
-                (oldViewer && !oldViewer.collapseDuplicates) ||
-                urls.indexOf(child.uri) == -1) {
-              urls.push({ uri: child.uri,
-                          isBookmark: this.nodeIsBookmark(child) });
-            }
-          }
-        }
-        if (!wasOpen)
-          aNode.containerOpen = false;
-      }
-      finally {
-        if (!wasOpen)
-          result.viewer = oldViewer;
-      }
+    var urls = [];
+    if (!this.nodeIsContainer(aNode))
+      return urls;
+
+    var root = this.getContainerNodeWithOptions(aNode, false, true);
+    var oldViewer = root.parentResult.viewer;
+    var wasOpen = root.containerOpen;
+    if (!wasOpen) {
+      root.parentResult.viewer = null;
+      root.containerOpen = true;
     }
 
+   for (var i = 0; i < root.childCount; ++i) {
+      var child = root.getChild(i);
+      if (this.nodeIsURI(child))
+        urls.push({uri: child.uri, isBookmark: this.nodeIsBookmark(child)});
+    }
+
+    if (!wasOpen) {
+      root.containerOpen = false;
+      root.parentResult.viewer = oldViewer;
+    }
     return urls;
   },
 
@@ -997,47 +1054,67 @@ var PlacesUtils = {
    *
    * @param aFile
    *        nsIFile of bookmarks in JSON format to be restored.
-   * @param aExcludeItems
-   *        Array of root item ids (ie: children of the places root)
-   *        to not delete when restoring.
    */
   restoreBookmarksFromJSONFile:
-  function PU_restoreBookmarksFromJSONFile(aFile, aExcludeItems) {
-    // open file stream
-    var stream = Cc["@mozilla.org/network/file-input-stream;1"].
-                 createInstance(Ci.nsIFileInputStream);
-    stream.init(aFile, 0x01, 0, 0);
-    var converted = Cc["@mozilla.org/intl/converter-input-stream;1"].
-                    createInstance(Ci.nsIConverterInputStream);
-    converted.init(stream, "UTF-8", 1024,
-                   Ci.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
+  function PU_restoreBookmarksFromJSONFile(aFile) {
+    var failed = false;
+    var obsServ = Cc["@mozilla.org/observer-service;1"].
+                  getService(Ci.nsIObserverService);
+    obsServ.notifyObservers(null,
+                            RESTORE_BEGIN_NSIOBSERVER_TOPIC,
+                            RESTORE_NSIOBSERVER_DATA);
 
-    // read in contents
-    var str = {};
-    var jsonStr = "";
-    while (converted.readString(4096, str) != 0)
-      jsonStr += str.value;
-    converted.close();
+    try {
+      // open file stream
+      var stream = Cc["@mozilla.org/network/file-input-stream;1"].
+                   createInstance(Ci.nsIFileInputStream);
+      stream.init(aFile, 0x01, 0, 0);
+      var converted = Cc["@mozilla.org/intl/converter-input-stream;1"].
+                      createInstance(Ci.nsIConverterInputStream);
+      converted.init(stream, "UTF-8", 8192,
+                     Ci.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
 
-    if (jsonStr.length == 0)
-      return; // empty file
+      // read in contents
+      var str = {};
+      var jsonStr = "";
+      while (converted.readString(8192, str) != 0)
+        jsonStr += str.value;
+      converted.close();
 
-    this.restoreBookmarksFromJSONString(jsonStr, true, aExcludeItems);
+      if (jsonStr.length == 0)
+        return; // empty file
+
+      this.restoreBookmarksFromJSONString(jsonStr, true);
+    }
+    catch (exc) {
+      failed = true;
+      obsServ.notifyObservers(null,
+                              RESTORE_FAILED_NSIOBSERVER_TOPIC,
+                              RESTORE_NSIOBSERVER_DATA);
+      Components.utils.reportError("Bookmarks JSON restore failed: " + exc);
+      throw exc;
+    }
+    finally {
+      if (!failed) {
+        obsServ.notifyObservers(null,
+                                RESTORE_SUCCESS_NSIOBSERVER_TOPIC,
+                                RESTORE_NSIOBSERVER_DATA);
+      }
+    }
   },
 
   /**
    * Import bookmarks from a JSON string.
+   * Note: any item annotated with "places/excludeFromBackup" won't be removed
+   *       before executing the restore.
    * 
    * @param aString
    *        JSON string of serialized bookmark data.
    * @param aReplace
    *        Boolean if true, replace existing bookmarks, else merge.
-   * @param aExcludeItems
-   *        Array of root item ids (ie: children of the places root)
-   *        to not delete when restoring.
    */
   restoreBookmarksFromJSONString:
-  function PU_restoreBookmarksFromJSONString(aString, aReplace, aExcludeItems) {
+  function PU_restoreBookmarksFromJSONString(aString, aReplace) {
     // convert string to JSON
     var nodes = this.unwrapNodes(aString, this.TYPE_X_MOZ_PLACE_CONTAINER);
 
@@ -1056,9 +1133,12 @@ var PlacesUtils = {
       nodes: nodes[0].children,
       runBatched: function restore_runBatched() {
         if (aReplace) {
-          var excludeItems = aExcludeItems || [];
+          // Get roots excluded from the backup, we will not remove them
+          // before restoring.
+          var excludeItems = this._utils.annotations
+                                 .getItemsWithAnnotation(EXCLUDE_FROM_BACKUP_ANNO, {});
           // delete existing children of the root node, excepting:
-          // 1. special folders: delete the child nodes 
+          // 1. special folders: delete the child nodes
           // 2. tags folder: untag via the tagging api
           var query = this._utils.history.getNewQuery();
           query.setFolders([this._utils.placesRootId], 1);
@@ -1080,11 +1160,26 @@ var PlacesUtils = {
               // remove tags via the tagging service
               var tags = this._utils.tagging.allTags;
               var uris = [];
+              var bogusTagContainer = false;
               for (let i in tags) {
-                var tagURIs = this._utils.tagging.getURIsForTag(tags[i]);
+                var tagURIs = [];
+                // skip empty tags since getURIsForTag would throw
+                if (tags[i])
+                  tagURIs = this._utils.tagging.getURIsForTag(tags[i]);
+
+                if (!tagURIs.length) {
+                  // This is a bogus tag container, empty tags should be removed
+                  // automatically, but this does not work if they contain some
+                  // not-uri node, so we remove them manually.
+                  // XXX this is a temporary workaround until we implement
+                  // preventive database maintenance in bug 431558.
+                  bogusTagContainer = true;
+                }
                 for (let j in tagURIs)
                   this._utils.tagging.untagURI(tagURIs[j], [tags[i]]);
               }
+              if (bogusTagContainer)
+                this._utils.bookmarks.removeFolderChildren(rootItemId);
             }
             else if ([this._utils.toolbarFolderId,
                       this._utils.unfiledBookmarksFolderId,
@@ -1123,7 +1218,10 @@ var PlacesUtils = {
             node.children.forEach(function(child) {
               var index = child.index;
               var [folders, searches] = this.importJSONNode(child, container, index);
-              folderIdMap = folderIdMap.concat(folders);
+              for (var i = 0; i < folders.length; i++) {
+                if (folders[i])
+                  folderIdMap[i] = folders[i];
+              }
               searchIds = searchIds.concat(searches);
             }, this);
           }
@@ -1166,9 +1264,14 @@ var PlacesUtils = {
     switch (aData.type) {
       case this.TYPE_X_MOZ_PLACE_CONTAINER:
         if (aContainer == PlacesUtils.bookmarks.tagsFolder) {
+          // node is a tag
           if (aData.children) {
             aData.children.forEach(function(aChild) {
-              this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              try {
+                this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              } catch (ex) {
+                // invalid tag child, skip it
+              }
             }, this);
             return [folderIdMap, searchIds];
           }
@@ -1178,30 +1281,41 @@ var PlacesUtils = {
           var feedURI = null;
           var siteURI = null;
           aData.annos = aData.annos.filter(function(aAnno) {
-            if (aAnno.name == LMANNO_FEEDURI) {
-              feedURI = this._uri(aAnno.value);
-              return false;
+            switch (aAnno.name) {
+              case LMANNO_FEEDURI:
+                feedURI = this._uri(aAnno.value);
+                return false;
+              case LMANNO_SITEURI:
+                siteURI = this._uri(aAnno.value);
+                return false;
+              case LMANNO_EXPIRATION:
+              case LMANNO_LOADING:
+              case LMANNO_LOADFAILED:
+                return false;
+              default:
+                return true;
             }
-            else if (aAnno.name == LMANNO_SITEURI) {
-              siteURI = this._uri(aAnno.value);
-              return false;
-            }
-            return true;
           }, this);
 
-          if (feedURI)
-            id = this.livemarks.createLivemark(aContainer, aData.title, siteURI, feedURI, aIndex);
+          if (feedURI) {
+            id = this.livemarks.createLivemarkFolderOnly(aContainer,
+                                                         aData.title,
+                                                         siteURI, feedURI,
+                                                         aIndex);
+          }
         }
         else {
           id = this.bookmarks.createFolder(aContainer, aData.title, aIndex);
-          folderIdMap.push([aData.id, id]);
+          folderIdMap[aData.id] = id;
           // process children
           if (aData.children) {
-            aData.children.every(function(aChild, aIndex) {
-              var [folderIds, searches] = this.importJSONNode(aChild, id, aIndex);
-              folderIdMap = folderIdMap.concat(folderIds);
+            aData.children.forEach(function(aChild, aIndex) {
+              var [folders, searches] = this.importJSONNode(aChild, id, aIndex);
+              for (var i = 0; i < folders.length; i++) {
+                if (folders[i])
+                  folderIdMap[i] = folders[i];
+              }
               searchIds = searchIds.concat(searches);
-              return true;
             }, this);
           }
         }
@@ -1217,20 +1331,23 @@ var PlacesUtils = {
         }
         if (aData.charset)
           this.history.setCharsetForURI(this._uri(aData.uri), aData.charset);
-        if (aData.uri.match(/^place:/))
+        if (aData.uri.substr(0, 6) == "place:")
           searchIds.push(id);
         break;
       case this.TYPE_X_MOZ_PLACE_SEPARATOR:
         id = this.bookmarks.insertSeparator(aContainer, aIndex);
         break;
       default:
+        // Unknown node type
     }
 
-    // set generic properties
+    // set generic properties, valid for all nodes
     if (id != -1) {
-      this.bookmarks.setItemDateAdded(id, aData.dateAdded);
-      this.bookmarks.setItemLastModified(id, aData.lastModified);
-      if (aData.annos)
+      if (aData.dateAdded)
+        this.bookmarks.setItemDateAdded(id, aData.dateAdded);
+      if (aData.lastModified)
+        this.bookmarks.setItemLastModified(id, aData.lastModified);
+      if (aData.annos && aData.annos.length)
         this.setAnnotationsForItem(id, aData.annos);
     }
 
@@ -1249,33 +1366,15 @@ var PlacesUtils = {
    *          returns the input URI unchanged.
    */
   _fixupQuery: function PU__fixupQuery(aQueryURI, aFolderIdMap) {
-    var queries = {};
-    var options = {};
-    this.history.queryStringToQueries(aQueryURI.spec, queries, {}, options);
-
-    var fixedQueries = [];
-    queries.value.forEach(function(aQuery) {
-      var folders = aQuery.getFolders({});
-
-      var newFolders = [];
-      aFolderIdMap.forEach(function(aMapping) {
-        if (folders.indexOf(aMapping[0]) != -1)
-          newFolders.push(aMapping[1]);
-      });
-
-      if (newFolders.length)
-        aQuery.setFolders(newFolders, newFolders.length);
-      fixedQueries.push(aQuery);
-    });
-
-    var stringURI = this.history.queriesToQueryString(fixedQueries,
-                                                      fixedQueries.length,
-                                                      options.value);
+    function convert(str, p1, offset, s) {
+      return "folder=" + aFolderIdMap[p1];
+    }
+    var stringURI = aQueryURI.spec.replace(/folder=([0-9]+)/g, convert);
     return this._uri(stringURI);
   },
 
   /**
-   * Serializes the given node (and all it's descendents) as JSON
+   * Serializes the given node (and all its descendents) as JSON
    * and writes the serialization to the given output stream.
    * 
    * @param   aNode
@@ -1328,7 +1427,7 @@ var PlacesUtils = {
               // When copying a read-only node, remove the read-only annotation.
               return false;
             }
-            return anno.name != "placesInternal/GUID";
+            return true;
           });
         } catch(ex) {
           LOG(ex);
@@ -1426,9 +1525,9 @@ var PlacesUtils = {
           var childNode = aSourceNode.getChild(i);
           if (aExcludeItems && aExcludeItems.indexOf(childNode.itemId) != -1)
             continue;
-          if (i != 0)
+          var written = serializeNodeToJSONStream(aSourceNode.getChild(i), i);
+          if (written && i < cc - 1)
             aStream.write(",", 1);
-          serializeNodeToJSONStream(aSourceNode.getChild(i), i);
         }
         if (!wasOpen)
           aSourceNode.containerOpen = false;
@@ -1449,33 +1548,69 @@ var PlacesUtils = {
 
       addGenericProperties(bNode, node);
 
-      if (self.nodeIsURI(bNode))
+      var parent = bNode.parent;
+      var grandParent = parent ? parent.parent : null;
+
+      if (self.nodeIsURI(bNode)) {
+        // Tag root accept only folder nodes
+        if (parent && parent.itemId == self.tagsFolderId)
+          return false;
+        // Check for url validity, since we can't halt while writing a backup.
+        // This will throw if we try to serialize an invalid url and it does
+        // not make sense saving a wrong or corrupt uri node.
+        try {
+          self._uri(bNode.uri);
+        } catch (ex) {
+          return false;
+        }
         addURIProperties(bNode, node);
-      else if (self.nodeIsContainer(bNode))
+      }
+      else if (self.nodeIsContainer(bNode)) {
+        // Tag containers accept only uri nodes
+        if (grandParent && grandParent.itemId == self.tagsFolderId)
+          return false;
         addContainerProperties(bNode, node);
-      else if (self.nodeIsSeparator(bNode))
+      }
+      else if (self.nodeIsSeparator(bNode)) {
+        // Tag root accept only folder nodes
+        // Tag containers accept only uri nodes
+        if ((parent && parent.itemId == self.tagsFolderId) ||
+            (grandParent && grandParent.itemId == self.tagsFolderId))
+          return false;
+
         addSeparatorProperties(bNode, node);
+      }
 
       if (!node.feedURI && node.type == self.TYPE_X_MOZ_PLACE_CONTAINER)
         writeComplexNode(aStream, node, bNode);
       else
         writeScalarNode(aStream, node);
+      return true;
     }
 
     // serialize to stream
     serializeNodeToJSONStream(aNode, null);
   },
 
-  // XXX testing serializers
+  /**
+   * Serialize a JS object to JSON
+   */
   toJSONString: function PU_toJSONString(aObj) {
     var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);
     return JSON.encode(aObj);
   },
 
   /**
+   * backupBookmarksToFile()
+   *
    * Serializes bookmarks using JSON, and writes to the supplied file.
+   * Note: any item that should not be backed up must be annotated with
+   *       "places/excludeFromBackup".
+   *
+   * @param aFile
+   *        nsIFile where to save JSON backup.
    */
-  backupBookmarksToFile: function PU_backupBookmarksToFile(aFile, aExcludeItems) {
+  backupBookmarksToFile: function PU_backupBookmarksToFile(aFile) {
     if (aFile.exists() && !aFile.isWritable())
       return; // XXX
 
@@ -1497,6 +1632,10 @@ var PlacesUtils = {
       }
     };
 
+    // Get itemIds to be exluded from the backup
+    var excludeItems = this.annotations
+                           .getItemsWithAnnotation(EXCLUDE_FROM_BACKUP_ANNO, {});
+
     // query places root
     var options = this.history.getNewQueryOptions();
     options.expandQueries = false;
@@ -1506,7 +1645,7 @@ var PlacesUtils = {
     result.root.containerOpen = true;
     // serialize as JSON, write to stream
     this.serializeNodeAsJSONToOutputStream(result.root, streamProxy,
-                                           false, false, aExcludeItems);
+                                           false, false, excludeItems);
     result.root.containerOpen = false;
 
     // close converter and stream
@@ -1519,6 +1658,8 @@ var PlacesUtils = {
    *
    * Creates a dated backup once a day in <profile>/bookmarkbackups.
    * Stores the bookmarks using JSON.
+   * Note: any item that should not be backed up must be annotated with
+   *       "places/excludeFromBackup".
    *
    * @param int aNumberOfBackups - the maximum number of backups to keep
    *
@@ -1542,17 +1683,26 @@ var PlacesUtils = {
     // Use YYYY-MM-DD (ISO 8601) as it doesn't contain illegal characters
     // and makes the alphabetical order of multiple backup files more useful.
     var date = new Date().toLocaleFormat("%Y-%m-%d");
-    var backupFilename = this.getFormattedString("bookmarksArchiveFilename", [date]);
+    var backupFilename = "bookmarks-" + date + ".json";
 
     var backupFile = null;
     if (!aForceArchive) {
       var backupFileNames = [];
       var backupFilenamePrefix = backupFilename.substr(0, backupFilename.indexOf("-"));
+
+      // Get the localized backup filename, to clear out
+      // old backups with a localized name (bug 445704).
+      var localizedFilename = this.getFormattedString("bookmarksArchiveFilename", [date]);
+      var localizedFilenamePrefix = localizedFilename.substr(0, localizedFilename.indexOf("-"));
+      var rx = new RegExp("^(bookmarks|" + localizedFilenamePrefix + ")-.+\.(json|html)");
+
       var entries = bookmarksBackupDir.directoryEntries;
       while (entries.hasMoreElements()) {
         var entry = entries.getNext().QueryInterface(Ci.nsIFile);
         var backupName = entry.leafName;
-        if (backupName.substr(0, backupFilenamePrefix.length) == backupFilenamePrefix) {
+        // A valid backup is any file that matches either the localized or
+        // not-localized filename (bug 445704).
+        if (backupName.match(rx)) {
           if (backupName == backupFilename)
             backupFile = entry;
           backupFileNames.push(backupName);
@@ -1626,5 +1776,13 @@ var PlacesUtils = {
     var backupFile = bookmarksBackupDir.clone();
     backupFile.append(filename);
     return backupFile;
+  },
+
+  /**
+   * Starts the database coherence check and executes update tasks on a timer,
+   * this method is called by browser.js in delayed startup.
+   */
+  startPlacesDBUtils: function PU_startPlacesDBUtils() {
+    Components.utils.import("resource://gre/modules/PlacesDBUtils.jsm");
   }
 };
